import type { Address, Provider, RpcRequest, RpcResponse, RpcSchema } from "ox";
import { z } from "zod";
export type SetPrimaryButton = (options: {
    text: string;
    loading?: boolean;
    disabled?: boolean;
    hidden?: boolean;
}) => void;
export type EthProviderRequest = Provider.RequestFn<RpcSchema.Default>;
export type AccountLocation = {
    placeId: string;
    /**
     * Human-readable string describing the location
     */
    description: string;
};
export type FrameLocationNotificationContext = {
    type: "notification";
    notification: {
        notificationId: string;
        title: string;
        body: string;
    };
};
export type FrameLocationContext = FrameLocationNotificationContext;
export type FrameContext = {
    user: {
        fid: number;
        username?: string;
        displayName?: string;
        /**
         * Profile image URL
         */
        pfpUrl?: string;
    };
    location?: FrameLocationContext;
};
export declare const notificationDetailsSchema: z.ZodObject<{
    url: z.ZodString;
    token: z.ZodString;
}, "strip", z.ZodTypeAny, {
    url: string;
    token: string;
}, {
    url: string;
    token: string;
}>;
export type FrameNotificationDetails = z.infer<typeof notificationDetailsSchema>;
export type AddFrameResult = {
    added: true;
    notificationDetails?: FrameNotificationDetails;
} | {
    added: false;
    reason: "invalid-domain-manifest" | "rejected-by-user";
};
export type AddFrame = () => Promise<AddFrameResult>;
export type FrameHost = {
    context: FrameContext;
    close: () => void;
    ready: () => void;
    openUrl: (url: string) => void;
    setPrimaryButton: SetPrimaryButton;
    ethProviderRequest: EthProviderRequest;
    ethProviderRequestV2: RpcTransport;
    addFrame: AddFrame;
};
export declare const eventSchema: z.ZodObject<{
    header: z.ZodString;
    payload: z.ZodString;
    signature: z.ZodString;
}, "strip", z.ZodTypeAny, {
    header: string;
    payload: string;
    signature: string;
}, {
    header: string;
    payload: string;
    signature: string;
}>;
export type EventSchema = z.infer<typeof eventSchema>;
export declare const eventHeaderSchema: z.ZodObject<{
    fid: z.ZodNumber;
    type: z.ZodLiteral<"app_key">;
    key: z.ZodString;
}, "strip", z.ZodTypeAny, {
    type: "app_key";
    fid: number;
    key: string;
}, {
    type: "app_key";
    fid: number;
    key: string;
}>;
export type EventHeader = z.infer<typeof eventHeaderSchema>;
export declare const eventFrameAddedPayloadSchema: z.ZodObject<{
    event: z.ZodLiteral<"frame-added">;
    notificationDetails: z.ZodOptional<z.ZodObject<{
        url: z.ZodString;
        token: z.ZodString;
    }, "strip", z.ZodTypeAny, {
        url: string;
        token: string;
    }, {
        url: string;
        token: string;
    }>>;
}, "strip", z.ZodTypeAny, {
    event: "frame-added";
    notificationDetails?: {
        url: string;
        token: string;
    } | undefined;
}, {
    event: "frame-added";
    notificationDetails?: {
        url: string;
        token: string;
    } | undefined;
}>;
export type EventFrameAddedPayload = z.infer<typeof eventFrameAddedPayloadSchema>;
export declare const eventFrameRemovedPayloadSchema: z.ZodObject<{
    event: z.ZodLiteral<"frame-removed">;
}, "strip", z.ZodTypeAny, {
    event: "frame-removed";
}, {
    event: "frame-removed";
}>;
export type EventFrameRemovedPayload = z.infer<typeof eventFrameRemovedPayloadSchema>;
export declare const eventNotificationsEnabledPayloadSchema: z.ZodObject<{
    event: z.ZodLiteral<"notifications-enabled">;
    notificationDetails: z.ZodObject<{
        url: z.ZodString;
        token: z.ZodString;
    }, "strip", z.ZodTypeAny, {
        url: string;
        token: string;
    }, {
        url: string;
        token: string;
    }>;
}, "strip", z.ZodTypeAny, {
    event: "notifications-enabled";
    notificationDetails: {
        url: string;
        token: string;
    };
}, {
    event: "notifications-enabled";
    notificationDetails: {
        url: string;
        token: string;
    };
}>;
export type EventNotificationsEnabledPayload = z.infer<typeof eventNotificationsEnabledPayloadSchema>;
export declare const notificationsDisabledPayloadSchema: z.ZodObject<{
    event: z.ZodLiteral<"notifications-disabled">;
}, "strip", z.ZodTypeAny, {
    event: "notifications-disabled";
}, {
    event: "notifications-disabled";
}>;
export type EventNotificationsDisabledPayload = z.infer<typeof notificationsDisabledPayloadSchema>;
export declare const eventPayloadSchema: z.ZodDiscriminatedUnion<"event", [z.ZodObject<{
    event: z.ZodLiteral<"frame-added">;
    notificationDetails: z.ZodOptional<z.ZodObject<{
        url: z.ZodString;
        token: z.ZodString;
    }, "strip", z.ZodTypeAny, {
        url: string;
        token: string;
    }, {
        url: string;
        token: string;
    }>>;
}, "strip", z.ZodTypeAny, {
    event: "frame-added";
    notificationDetails?: {
        url: string;
        token: string;
    } | undefined;
}, {
    event: "frame-added";
    notificationDetails?: {
        url: string;
        token: string;
    } | undefined;
}>, z.ZodObject<{
    event: z.ZodLiteral<"frame-removed">;
}, "strip", z.ZodTypeAny, {
    event: "frame-removed";
}, {
    event: "frame-removed";
}>, z.ZodObject<{
    event: z.ZodLiteral<"notifications-enabled">;
    notificationDetails: z.ZodObject<{
        url: z.ZodString;
        token: z.ZodString;
    }, "strip", z.ZodTypeAny, {
        url: string;
        token: string;
    }, {
        url: string;
        token: string;
    }>;
}, "strip", z.ZodTypeAny, {
    event: "notifications-enabled";
    notificationDetails: {
        url: string;
        token: string;
    };
}, {
    event: "notifications-enabled";
    notificationDetails: {
        url: string;
        token: string;
    };
}>, z.ZodObject<{
    event: z.ZodLiteral<"notifications-disabled">;
}, "strip", z.ZodTypeAny, {
    event: "notifications-disabled";
}, {
    event: "notifications-disabled";
}>]>;
export type FrameEvent = z.infer<typeof eventPayloadSchema>;
export declare const sendNotificationRequestSchema: z.ZodObject<{
    notificationId: z.ZodString;
    title: z.ZodString;
    body: z.ZodString;
    targetUrl: z.ZodString;
    tokens: z.ZodArray<z.ZodString, "many">;
}, "strip", z.ZodTypeAny, {
    notificationId: string;
    title: string;
    body: string;
    targetUrl: string;
    tokens: string[];
}, {
    notificationId: string;
    title: string;
    body: string;
    targetUrl: string;
    tokens: string[];
}>;
export type SendNotificationRequest = z.infer<typeof sendNotificationRequestSchema>;
export declare const sendNotificationResponseSchema: z.ZodObject<{
    result: z.ZodObject<{
        successfulTokens: z.ZodArray<z.ZodString, "many">;
        invalidTokens: z.ZodArray<z.ZodString, "many">;
        rateLimitedTokens: z.ZodArray<z.ZodString, "many">;
    }, "strip", z.ZodTypeAny, {
        successfulTokens: string[];
        invalidTokens: string[];
        rateLimitedTokens: string[];
    }, {
        successfulTokens: string[];
        invalidTokens: string[];
        rateLimitedTokens: string[];
    }>;
}, "strip", z.ZodTypeAny, {
    result: {
        successfulTokens: string[];
        invalidTokens: string[];
        rateLimitedTokens: string[];
    };
}, {
    result: {
        successfulTokens: string[];
        invalidTokens: string[];
        rateLimitedTokens: string[];
    };
}>;
export type SendNotificationResponse = z.infer<typeof sendNotificationResponseSchema>;
export type RpcTransport = (request: RpcRequest.RpcRequest) => Promise<RpcResponse.RpcResponse>;
export type ProviderRpcError = {
    code: number;
    details?: string;
    message?: string;
};
export type EthProviderWireEvent = {
    event: "accountsChanged";
    params: [readonly Address.Address[]];
} | {
    event: "chainChanged";
    params: [string];
} | {
    event: "connect";
    params: [Provider.ConnectInfo];
} | {
    event: "disconnect";
    params: [ProviderRpcError];
} | {
    event: "message";
    params: [Provider.Message];
};
export type EmitEthProvider = <event extends EthProviderWireEvent['event']>(event: event, params: Extract<EthProviderWireEvent, {
    event: event;
}>['params']) => void;
